#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <git2.h>
#include "repo.h"
#include "md5.h"

static int add_file(const char *name, const char *path);
static void calc_chksum(FILE *fp, uint32_t *sum);
static int proc_tree(git_repository *repo, git_tree *tree, const char *path);

static struct repo_file *files;
static int num_files, max_files;
static const char *repo_path;

int repo_init(const char *path)
{
	int res = -1;
	git_repository *repo = 0;
	git_object *tree;

	git_libgit2_init();

	if(git_repository_open_ext(&repo, path, GIT_REPOSITORY_OPEN_NO_SEARCH, 0) != 0) {
		fprintf(stderr, "repo_init: failed to open repository: %s: %s\n", path,
				git_error_last()->message);
		goto end;
	}

	files = 0;
	num_files = max_files = 0;
	repo_path = path;

	if(git_revparse_single(&tree, repo, "HEAD^{tree}") != 0) {
		fprintf(stderr, "repo_init: %s\n", git_error_last()->message);
		goto end;
	}
	if(proc_tree(repo, (git_tree*)tree, 0) == -1) {
		goto end;
	}
	git_object_free(tree);

	res = 0;
end:
	if(res != 0) {
		free(files);
		files = 0;
		num_files = max_files = 0;
	}
	if(repo) git_repository_free(repo);
	git_libgit2_shutdown();
	return res;
}

void repo_cleanup(void)
{
	int i;
	for(i=0; i<num_files; i++) {
		free(files[i].fullpath);
	}
	free(files);
	files = 0;
	num_files = max_files = 0;
}

int repo_num_files(void)
{
	return num_files;
}

struct repo_file *repo_file(int idx)
{
	if(idx < 0 || idx >= num_files) return 0;
	return files + idx;
}

static int add_file(const char *name, const char *path)
{
	struct repo_file *file;
	int pathlen;
	struct stat st;
	FILE *fp;

	if(num_files >= max_files) {
		int newmax = max_files ? max_files * 2 : 16;
		void *newfiles = realloc(files, newmax * sizeof *files);
		if(!newfiles) {
			fprintf(stderr, "repo_init: failed to add file\n");
			return -1;
		}
		files = newfiles;
		max_files = newmax;
	}

	pathlen = strlen(repo_path) + strlen(name) + 1;
	if(path) {
		pathlen += strlen(path) + 1;
	}

	file = files + num_files;
	if(!(file->fullpath = malloc(pathlen + 1))) {
		fprintf(stderr, "repo_init: failed to allocate file path buffer\n");
		return -1;
	}
	if(path) {
		sprintf(file->fullpath, "%s/%s/%s", repo_path, path, name);
	} else {
		sprintf(file->fullpath, "%s/%s", repo_path, name);
	}
	file->path = file->fullpath + strlen(repo_path) + 1;

	/* attempt to open the file */
	if(!(fp = fopen(file->fullpath, "rb"))) {
		fprintf(stderr, "repo_init: failed to open file: %s: %s\n", file->fullpath, strerror(errno));
		free(file->fullpath);
		return -1;
	}

	/* get size & modification time */
	fstat(fileno(fp), &st);
	file->size = st.st_size;
	file->mtime = st.st_mtime;

	/* calculate md5 checksum */
	calc_chksum(fp, file->chksum);

	fclose(fp);

	num_files++;
	return 0;
}

static void calc_chksum(FILE *fp, uint32_t *sum)
{
	static char buf[4096];
	int sz;
	struct md5_state md;

	md5_begin(&md);
	while((sz = fread(buf, 1, sizeof buf, fp)) > 0) {
		md5_msg(&md, buf, sz);
	}
	md5_end(&md);

	memcpy(sum, md.sum, sizeof md.sum);
}

static int proc_tree(git_repository *repo, git_tree *tree, const char *path)
{
	int i, count = git_tree_entrycount(tree);
	const git_tree_entry *ent;
	git_object_t type;
	git_object *obj;
	const char *name;

	for(i=0; i<count; i++) {
		ent = git_tree_entry_byindex(tree, i);
		type = git_tree_entry_type(ent);
		name = git_tree_entry_name(ent);

		switch(type) {
		case GIT_OBJECT_BLOB:
			if(add_file(name, path) == -1) {
				return -1;
			}
			break;

		case GIT_OBJECT_TREE:
			git_tree_entry_to_object(&obj, repo, ent);
			if(path) {
				char *pathbuf = malloc(strlen(path) + strlen(name) + 2);
				if(!pathbuf) {
					fprintf(stderr, "repo_init: failed to allocate memory\n");
					return -1;
				}
				sprintf(pathbuf, "%s/%s", path, name);
				proc_tree(repo, (git_tree*)obj, pathbuf);
				free(pathbuf);
			} else {
				proc_tree(repo, (git_tree*)obj, name);
			}
			git_object_free(obj);
			break;

		default:
			break;
		}
	}

	return 0;
}
